/*
Module : ClockCtrl.cpp
Purpose: Implementation for an MFC GUI control which implements
         an analog clock
Created: PJN / 01-02-2000
History: PJN / 16-02-2000 1. Removed noticeable flicker which sometimes occurred 
                             when the seconds hand moved
         PJN / 27-07-2001 1. Fixed a drawing glitch which left the old hour hand 
                             being displayed every 12 minutes
                          2. Made most of the functions virtual to allow easier
                          customization.
         PJN / 01-08-2001 1. Second hand can now be updated with sub second precision
                          with the "SetSubSecondPrecision" method
         PJN / 23-03-2001 1. Updated Copyright details.
                          2. Fixed a bug where PreSubclassWindow was not calling 
                          RecalcLayout before a redraw was occurring.
         PJN / 31-12-2016 1. Updated copyright details.
                          2. Replaced all occurrences of NULL with nullptr throughout the codebase
                          3. Updated the code to clean compile on VC 2010 to VC 2015. The code now
                          requires VC 2010 at a minimum.
                          4. Added SAL annotations to all the code
         PJN / 29-12-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes 
                          mean that the code will now only compile on VC 2017 or later.
         PJN / 03-06-2019 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2019
         PJN / 28-09-2019 1. Fixed a number of compiler warnings when the code is compiled with VS 2019 Preview
         PJN / 26-12-2019 1. Fixed various Clang-Tidy static code analysis warnings in the code.
         PJN / 04-04-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 07-05-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations.

Copyright (c) 2000 - 2022 by PJ Naughter  (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/

//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#define _USE_MATH_DEFINES
#include <math.h> //NOLINT(modernize-deprecated-headers)
#include "ClockCtrl.h"


//////////////////// Macros / Defines /////////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif


//////////////////// Implementation ///////////////////////////////////////////

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CClockCtrl, CStatic) //NOLINT(modernize-avoid-c-arrays)
  ON_WM_PAINT()
  ON_WM_SIZE()
  ON_WM_TIMER()
  ON_WM_DESTROY()
  ON_WM_ERASEBKGND()
END_MESSAGE_MAP()

CClockCtrl::CClockCtrl() noexcept : m_ColorHourMinutesHand{GetSysColor(COLOR_HIGHLIGHT)},
                                    m_ColorBackground{GetSysColor(COLOR_BTNFACE)},
                                    m_nXRadius{-1},
                                    m_nYRadius{-1},
                                    m_nPointWidth{-1},
                                    m_nHourPosAsMinutes{0},
                                    m_nHour{0},
                                    m_nMinute{0},
                                    m_nSecond{0},
                                    m_nMillisecond{0},
                                    m_nTimerID{0},
                                    m_bEnableRealtime{TRUE},
                                    m_bShowHours{true},
                                    m_bShowMinutes{true},
                                    m_bShowSeconds{true},
                                    m_bShowPoints{true},
                                    m_bShowMinorPoints{true},
                                    m_b3dPoints{true},
                                    m_bSubSecondPrecision{true}
{
}

void CClockCtrl::PreSubclassWindow()
{
  //Let the parent class do its thing
  __super::PreSubclassWindow();

  //Setup the timer if we are displaying realtime values
  if (m_bEnableRealtime)
  {
    RecalcLayout();
    SetRealTime(true);
  }
}

#pragma warning(suppress: 26434)
void CClockCtrl::OnPaint()
{
  //Force a recalc if not initialized
  if (m_nXRadius == -1)
    RecalcLayout();

  //device context for painting
  CPaintDC dc{this};

  //Setup the brush we are going to use
  const COLORREF colorPoints{GetSysColor(COLOR_BTNFACE)};
  CBrush PointBrush{colorPoints};
  CBrush* pOldBrush{dc.SelectObject(&PointBrush)};

  //Draw the points going around the clock
  if (m_bShowPoints)
  {
    for (int nMinute=0; nMinute<60; nMinute++)
    {
      //Work out the face length which is different 
      //depending on where we are on the ellipse edge
      const CPoint p{ComputePoint(nMinute, 0.9)};
      if ((nMinute % 5) == 0)
        DrawFacePoint(dc, p, true);
      else if (m_bShowMinorPoints)
        DrawFacePoint(dc, p, false);
    }
  }
  //Restore the DC
  dc.SelectObject(pOldBrush);

  //Then draw all the hands
  DrawHands(dc, m_nHour, m_nMinute, m_nSecond, m_nMillisecond, true);
}

void CClockCtrl::DrawHands(_In_ CDC& dc, _In_ int nHour, _In_ int nMinute, _In_ int nSecond, _In_ int nMillisecond, _In_ bool bDrawAll)
{
  //Validate our parameters
  ASSERT((nHour >= 0) && (nHour <= 12));
  ASSERT((nMinute >= 0) && (nMinute <= 59));
  ASSERT((nSecond >= 0) && (nSecond <= 59));
  ASSERT((nMillisecond >= 0) && (nMillisecond <= 999));

  //Setup the brush we are going to use
  CBrush PointBrush{m_ColorHourMinutesHand};
  CBrush* pOldBrush{dc.SelectObject(&PointBrush)};

  const int nHourPosAsMinutes{(m_nHour * 5) + (m_nMinute / 12)};
  if (bDrawAll)
  {
    //Draw the hour hand
    if (m_bShowHours)
      DrawHand(dc, nHourPosAsMinutes, HandType::HOUR, true);

    //Draw the minute hand
    if (m_bShowMinutes)
      DrawHand(dc, nMinute, HandType::MINUTE, true);

    //Draw the seconds hand
    if (m_bShowSeconds)
    {
      if (m_bSubSecondPrecision)
        DrawSecondHand(dc, nSecond, nMillisecond);
      else
        DrawHand(dc, nSecond, HandType::SECOND, true);
    }
  }
  else
  {
    //Should we do a complete redraw
    const bool bCompleteRedraw{(m_nHour != nHour) || (m_nMinute != nMinute) || (m_nHourPosAsMinutes != nHourPosAsMinutes)};
    if (bCompleteRedraw)
    {
      //Erase the seconds hand
      if (m_bShowSeconds)
      {
        if (m_bSubSecondPrecision)
          DrawSecondHand(dc, m_nSecond, m_nMillisecond);
        else
          DrawHand(dc, m_nSecond, HandType::SECOND, false);
      }

      //Erase the hour hand
      if (m_bShowHours)
        DrawHand(dc, m_nHourPosAsMinutes, HandType::HOUR, false);

      //Erase the minute hand
      if (m_bShowMinutes)
        DrawHand(dc, m_nMinute, HandType::MINUTE, false);

      //Draw the hour hand
      if (m_bShowHours)
        DrawHand(dc, nHourPosAsMinutes, HandType::HOUR, true);

      //Draw the minute hand
      if (m_bShowMinutes)
        DrawHand(dc, nMinute, HandType::MINUTE, true);

      //Draw the seconds hand
      if (m_bShowSeconds)
      {
        if (m_bSubSecondPrecision)
          DrawSecondHand(dc, nSecond, nMillisecond);
        else
          DrawHand(dc, nSecond, HandType::SECOND, true);
      }
    }
    else
    {
      //Erase the seconds hand
      if (m_bShowSeconds)
      {
        if (m_bSubSecondPrecision)
          DrawSecondHand(dc, m_nSecond, m_nMillisecond);
        else
          DrawHand(dc, m_nSecond, HandType::SECOND, false);
      }

      //Draw the seconds hand
      if (m_bShowSeconds)
      {
        if (m_bSubSecondPrecision)
          DrawSecondHand(dc, nSecond, nMillisecond);
        else
          DrawHand(dc, nSecond, HandType::SECOND, true);
      }
    }
  }

  //Store away the new values
  m_nHour = nHour;
  m_nHourPosAsMinutes = nHourPosAsMinutes;
  m_nMinute = nMinute;
  m_nSecond = nSecond;
  m_nMillisecond = nMillisecond;

  //Restore the DC
  dc.SelectObject(pOldBrush);
}

void CClockCtrl::SetTime(_In_ int nHour, _In_ int nMinute, _In_ int nSecond, _In_ int nMillisecond)
{
  //Validate our parameters
  ASSERT(m_hWnd);
  ASSERT((nHour >= 0) && (nHour <= 12));
  ASSERT((nMinute >= 0) && (nMinute <= 59));
  ASSERT((nSecond >= 0) && (nSecond <= 59));
  ASSERT((nMillisecond >= 0) && (nMillisecond <= 999));

  //Do the updated drawing of the hands
  CClientDC dc{this};
  DrawHands(dc, nHour, nMinute, nSecond, nMillisecond, false);
}

double CClockCtrl::MinuteToRadian(_In_ double minute) noexcept
{
  return (minute - 15)*M_PI/30;
}

CPoint CClockCtrl::ComputeSecondsPoint(_In_ double Minute, _In_ double ratio) noexcept
{
  const double angle{MinuteToRadian(Minute)};
  CPoint point;
  point.x = m_MiddlePoint.x + static_cast<int>(m_nXRadius*ratio*cos(angle) + 0.5);
  point.y = m_MiddlePoint.y + static_cast<int>(m_nYRadius*ratio*sin(angle) + 0.5);
  return point;
}

CPoint CClockCtrl::ComputePoint(_In_ int nMinute, _In_ double ratio) noexcept
{
  const double angle{MinuteToRadian(nMinute)};
  CPoint point;
  point.x = m_MiddlePoint.x + static_cast<int>(m_nXRadius*ratio*cos(angle) + 0.5);
  point.y = m_MiddlePoint.y + static_cast<int>(m_nYRadius*ratio*sin(angle) + 0.5);
  return point;
}

void CClockCtrl::DrawFacePoint(_In_ CDC& dc, _In_ const CPoint& point, _In_ bool bMajor)
{
  //Work out the size of the point rectangle
  CRect rPoint{point, point};
  if (bMajor)
  {
    const int nPointRadius{(m_nPointWidth / 2) + 1};
    rPoint.InflateRect(nPointRadius, nPointRadius);

    //Do the actual point drawing
    dc.Rectangle(&rPoint);
    if (m_b3dPoints)
      dc.Draw3dRect(&rPoint, GetSysColor(COLOR_BTNHIGHLIGHT), GetSysColor(COLOR_BTNSHADOW));
  }
  else
  {
    if (m_bShowMinorPoints)
    {
      rPoint.InflateRect(1, 1);

      //Do the actual point drawing
      dc.Draw3dRect(&rPoint, GetSysColor(COLOR_BTNHIGHLIGHT), GetSysColor(COLOR_BTNSHADOW));
    }
  }
}

void CClockCtrl::DrawSecondHand(_In_ CDC& dc, _In_ int nSecond, _In_ int nMilliSecond)
{
  std::array<CPoint, 2> handPoints;
  GetSecondsHandPoints(nSecond + (nMilliSecond/1000.0), handPoints);

  //Draw the hand
  const int nOldRop{dc.SetROP2(R2_NOTXORPEN)};
#pragma warning(suppress: 26446)
  dc.MoveTo(handPoints[0]);
#pragma warning(suppress: 26446)
  dc.LineTo(handPoints[1]);
  dc.SetROP2(nOldRop);
}

void CClockCtrl::DrawHand(_In_ CDC& dc, _In_ int nMinute, _In_ HandType type, _In_ bool bDraw)
{
  //Calculate the point positions of the hand
  std::array<CPoint, 4> handPoints;
  GetHandPoints(nMinute, type, handPoints);

  if (type == HandType::SECOND)
  {
    //Draw the hand
    const int nOldRop{dc.SetROP2(R2_NOTXORPEN)};
#pragma warning(suppress: 26446)
    dc.MoveTo(handPoints[0]);
#pragma warning(suppress: 26446)
    dc.LineTo(handPoints[1]);
    dc.SetROP2(nOldRop);
  }
  else
  {
    //Pick the correct colors to be used
    COLORREF colorBrush{RGB(0, 0, 0)};
    COLORREF colorPen{RGB(0, 0, 0)};
    if (bDraw)
    {
      colorBrush = m_ColorHourMinutesHand;
      colorPen = RGB(0, 0, 0);
    }
    else
    {
      colorBrush = m_ColorBackground;
      colorPen = m_ColorBackground;
    }

    //Setup the DC
    CBrush brushHand{colorBrush};
    CPen penHand{PS_SOLID, 1, colorPen};
    CBrush* pOldBrush{dc.SelectObject(&brushHand)};
    CPen* pOldPen{dc.SelectObject(&penHand)};

    //Draw the hand
    dc.Polygon(handPoints.data(), static_cast<int>(handPoints.size()));

    //Restore the DC
    dc.SelectObject(pOldPen);
    dc.SelectObject(pOldBrush);
  }
}

void CClockCtrl::GetSecondsHandPoints(_In_ double Value, _Inout_ std::array<CPoint, 2>& points) noexcept
{
  //Second hand goes out to 80% of the radius
  constexpr const double ratio{0.8};

#pragma warning(suppress: 26446)
  points[0] = m_MiddlePoint;
#pragma warning(suppress: 26446)
  points[1] = ComputeSecondsPoint(Value, ratio);
}

#pragma warning(suppress: 26440)
void CClockCtrl::GetHandPoints(_In_ int nValue, _In_ HandType type, _Inout_ std::array<CPoint, 4>& points)
{
  double ratio{0};
  switch (type)
  {
    case HandType::HOUR:
    {
      //hour hand goes out to 50% of the radius
      ratio = 0.5;
      break;
    }
    case HandType::MINUTE:
    {
      //Minute hand goes out to 70% of the radius
      ratio = 0.7;
      break;
    }
    case HandType::SECOND:
    {
      //Second hand goes out to 80% of the radius
      ratio = 0.8;
      break;
    }
    default:
    {
      ASSERT(false);
      break;
    }
  }

  if (type == HandType::SECOND)
  {
#pragma warning(suppress: 26446)
    points[0] = m_MiddlePoint;
#pragma warning(suppress: 26446)
    points[1] = ComputePoint(nValue, ratio);
  }
  else
  {
    //Compute the face points, First point is the back side,
    //second point is the right, third point is the top,
    //and the fourth is left.
#pragma warning(suppress: 26446)
    points[0] = ComputePoint(nValue + 30, 0.1);
#pragma warning(suppress: 26446)
    points[1] = ComputePoint(nValue + 15, 0.05);
#pragma warning(suppress: 26446)
    points[2] = ComputePoint(nValue, ratio);
#pragma warning(suppress: 26446)
    points[3] = ComputePoint(nValue - 15, 0.05);
  }
}

#pragma warning(suppress: 26434)
void CClockCtrl::OnSize(UINT nType, int cx, int cy)
{
  //Let the base class do its thing
  __super::OnSize(nType, cx, cy);

  //Force a recalc of the internal control metrics
  RecalcLayout();
}

void CClockCtrl::RecalcLayout()
{
  //Get the window rect
  CRect rClient;
  GetClientRect(&rClient);

  //Calculate the middle point
  m_MiddlePoint.x = rClient.Width() / 2;
  m_MiddlePoint.y = rClient.Height() / 2;

  //Calculate the radius
  m_nXRadius = m_MiddlePoint.x;
  m_nYRadius = m_MiddlePoint.y;

  //Calculate the width of the marks going around the clock
  m_nPointWidth = min(m_nXRadius, m_nYRadius) / 20;
  if (m_nPointWidth < 2)
    m_nPointWidth = 2;

  Invalidate(TRUE);
}

#pragma warning(suppress: 26434)
void CClockCtrl::OnTimer(UINT_PTR nIDEvent)
{
  //Let the base class do its thing
  __super::OnTimer(nIDEvent);

  if (m_nTimerID == nIDEvent)
  {
    //Get the local time and force a redraw
    SYSTEMTIME st;
    GetLocalTime(&st);
    SetTime(st.wHour % 12, st.wMinute, st.wSecond, st.wMilliseconds);
  }
}

void CClockCtrl::SetRealTime(_In_ bool bRealTime)
{
  m_bEnableRealtime = bRealTime;

  //Force an explicit refresh of the current state before we go realtime
  if (m_bEnableRealtime && m_hWnd)
  {
    CClientDC dc{this};

    //Redraw the background
    CRect rClient;
    GetClientRect(&rClient);
    dc.FillSolidRect(rClient, m_ColorBackground);

    //Redraw the hand
    DrawHands(dc, m_nHour, m_nMinute, m_nSecond, m_nMillisecond, true);
  }

  if (bRealTime)
  {
    if (m_nTimerID == 0)
    {
      SYSTEMTIME st;
      GetLocalTime(&st);
      SetTime(st.wHour % 12, st.wMinute, st.wSecond, st.wMilliseconds);

      if (m_bSubSecondPrecision)
        m_nTimerID = SetTimer(1, 50, nullptr); //Use a higher accuracy timer if we are doing sub second drawing
      else
        m_nTimerID = SetTimer(1, 100, nullptr);
    }
  }
  else
  {
    if (m_nTimerID)
    {
      KillTimer(m_nTimerID);
      m_nTimerID = 0;
    }
  }
}

bool CClockCtrl::GetRealTime() const noexcept
{
  return (m_nTimerID != 0);
}

#pragma warning(suppress: 26434)
void CClockCtrl::OnDestroy()
{
  //Turn off timers if we are using them
  SetRealTime(false);

  //Let the base class do its thing
  __super::OnDestroy();
}

void CClockCtrl::SetShowHours(_In_ bool bShowHours)
{
  if (bShowHours != m_bShowHours)
  {
    m_bShowHours = bShowHours;
    if (GetSafeHwnd())
      Invalidate();
  }
}

void CClockCtrl::SetShowMinutes(_In_ bool bShowMinutes)
{
  if (bShowMinutes != m_bShowMinutes)
  {
    m_bShowMinutes = bShowMinutes;
    if (GetSafeHwnd())
      Invalidate();
  }
}

void CClockCtrl::SetShowSeconds(_In_ bool bShowSeconds)
{
  if (bShowSeconds != m_bShowSeconds)
  {
    m_bShowSeconds = bShowSeconds;
    if (GetSafeHwnd())
      Invalidate();
  }
}

void CClockCtrl::SetShowPoints(_In_ bool bShowPoints)
{
  if (bShowPoints != m_bShowPoints)
  {
    m_bShowPoints = bShowPoints;
    if (GetSafeHwnd())
      Invalidate();
  }
}

void CClockCtrl::SetShowMinorPoints(_In_ bool bShowMinorPoints)
{
  if (bShowMinorPoints != m_bShowMinorPoints)
  {
    m_bShowMinorPoints = bShowMinorPoints;
    if (GetSafeHwnd())
      Invalidate();
  }
}

void CClockCtrl::SetHourMinutesHandColor(_In_ COLORREF color)
{
  if (color != m_ColorHourMinutesHand)
  {
    m_ColorHourMinutesHand = color;
    if (GetSafeHwnd())
      Invalidate();
  }
}

void CClockCtrl::SetBackgroundColor(_In_ COLORREF color)
{
  if (color != m_ColorBackground)
  {
    m_ColorBackground = color;
    if (GetSafeHwnd())
      Invalidate();
  }
}

void CClockCtrl::Set3dPoints(_In_ bool b3dPoints)
{
  if (b3dPoints != m_b3dPoints)
  {
    m_b3dPoints = b3dPoints;
    if (GetSafeHwnd())
      Invalidate();
  }
}

void CClockCtrl::SetSubSecondPrecision(_In_ bool bSubSecondPrecision)
{
  if (bSubSecondPrecision != m_bSubSecondPrecision)
  {
    m_bSubSecondPrecision = bSubSecondPrecision;
    if (GetSafeHwnd())
      Invalidate();
  }
}

#pragma warning(suppress: 26429 26434)
BOOL CClockCtrl::OnEraseBkgnd(CDC* pDC)
{
  CRect rClient;
  GetClientRect(&rClient);
  pDC->FillSolidRect(rClient, m_ColorBackground);
  return TRUE;
}
