/*
Module : ClockCtrl.h
Purpose: Defines the interface for an MFC GUI control which implements
         an analog clock
Created: PJN / 01-02-2000

Copyright (c) 2000 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#pragma once

#ifndef _CLOCKCTRL_H__
#define _CLOCKCTRL_H__

#ifndef CCLOCKCTRL_EXT_CLASS
#define CCLOCKCTRL_EXT_CLASS
#endif //#ifndef CCLOCKCTRL_EXT_CLASS

#ifndef _ARRAY_
#pragma message("To avoid this message, please put array in your pre compiled header (usually stdafx.h)")
#include <array>
#endif //#ifndef _ARRAY_


//////////////////// Classes //////////////////////////////////////////////////

class CCLOCKCTRL_EXT_CLASS CClockCtrl : public CStatic
{
public:
//Constructors / Destructors
  CClockCtrl() noexcept;
  CClockCtrl(const CClockCtrl&) = delete;
  CClockCtrl(CClockCtrl&&) = delete;
  ~CClockCtrl() = default;

//Methods
  CClockCtrl& operator=(const CClockCtrl&) = delete;
  CClockCtrl& operator=(CClockCtrl&&) = delete;
  int GetHour() const noexcept { return m_nHour; };
  int GetMinute() const noexcept { return m_nMinute; };
  int GetSecond() const noexcept { return m_nSecond; };
  bool GetRealTime() const noexcept;
  bool GetShowHours() const noexcept { return m_bShowHours; };
  bool GetShowMinutes() const noexcept { return m_bShowMinutes; };
  bool GetShowSeconds() const noexcept { return m_bShowSeconds; };
  bool GetShowPoints() const noexcept { return m_bShowPoints; };
  bool GetShowMinorPoints() const noexcept { return m_bShowMinorPoints; };
  COLORREF GetHourMinutesHandColor() const noexcept { return m_ColorHourMinutesHand; };
  COLORREF GetBackgroundColor() const noexcept { return m_ColorBackground; };
  bool Get3dPoints() const noexcept { return m_b3dPoints; };
  bool GetSubSecondPrecision() const noexcept { return m_bSubSecondPrecision; };
  void SetTime(_In_ int nHour, _In_ int nMinute, _In_ int nSecond, _In_ int nMillisecond = 0);
  void SetRealTime(_In_ bool bRealTime);
  void SetShowHours(_In_ bool bShowHours);
  void SetShowMinutes(_In_ bool bShowMinutes);
  void SetShowSeconds(_In_ bool bShowSeconds);
  void SetShowPoints(_In_ bool bShowPoints);
  void SetShowMinorPoints(_In_ bool bShowMinorPoints);
  void SetHourMinutesHandColor(_In_ COLORREF color);
  void SetBackgroundColor(_In_ COLORREF color);
  void Set3dPoints(_In_ bool b3dPoints);
  void SetSubSecondPrecision(_In_ bool bSubSecondPrecision);

protected:
//Enums
  enum class HandType
  {
    HOUR,
    MINUTE,
    SECOND
  };

//Methods
  void PreSubclassWindow() override;
  void RecalcLayout();
  double MinuteToRadian(_In_ double minute) noexcept;
  virtual CPoint ComputePoint(_In_ int nMinute, _In_ double ratio) noexcept;
  virtual CPoint ComputeSecondsPoint(_In_ double Minute, _In_ double ratio) noexcept;
  virtual void DrawFacePoint(_In_ CDC& dc, _In_ const CPoint& point, _In_ bool bMajor);
  virtual void DrawHands(_In_ CDC& dc, _In_ int nHour, _In_ int nMinute, _In_ int nSecond, _In_ int nMillisecond, _In_ bool bDrawAll);
  virtual void DrawHand(_In_ CDC& dc, _In_ int nMinute, _In_ HandType type, _In_ bool bDraw);
  virtual void GetHandPoints(_In_ int nValue, _In_ HandType type, _Inout_ std::array<CPoint, 4>& points);
  virtual void GetSecondsHandPoints(_In_ double Value, _Inout_ std::array<CPoint, 2>& points) noexcept;
  virtual void DrawSecondHand(_In_ CDC& dc, _In_ int nSecond, _In_ int nMilliSecond);

//Message handlers
  afx_msg void OnPaint();
  afx_msg void OnSize(UINT nType, int cx, int cy);
  afx_msg void OnTimer(UINT_PTR nIDEvent);
  afx_msg void OnDestroy();
  afx_msg BOOL OnEraseBkgnd(CDC* pDC);

//Member variables
  COLORREF m_ColorHourMinutesHand;
  COLORREF m_ColorBackground;
  CPoint m_MiddlePoint;
  int m_nXRadius;
  int m_nYRadius;
  int m_nPointWidth;
  int m_nHourPosAsMinutes;
  int m_nHour;
  int m_nMinute;
  int m_nSecond;
  int m_nMillisecond;
  UINT_PTR m_nTimerID;
  bool m_bEnableRealtime;
  bool m_bShowHours;
  bool m_bShowMinutes;
  bool m_bShowSeconds;
  bool m_bShowPoints;
  bool m_bShowMinorPoints;
  bool m_b3dPoints;
  bool m_bSubSecondPrecision;

  DECLARE_MESSAGE_MAP()
};

#endif //#ifndef _CLOCKCTRL_H__
